#include <SoftwareSerial.h>
#include <Servo.h>
//SETTINGS
#define PINGEWINNTRX 4            // Serielle Kommunikation mit 4-Gewinnt-Bot
#define PINGEWINNTTX 11           // Serielle Kommunikation mit 4-Gewinnt-Bot
#define PINGEWINNTSTART 2         // Kommunikation mit 4-Gewinnt-Bot: 1, wenn ein Spiel laufen soll, 0 sonst
#define PINSONGSTART 5            // Kommunikation mit Song-Bot: 1, wenn ein Lied gespielt werden soll, 0 sonst
#define PIN_SERVO 6               // Pin fuer den Servo zum leeren des Spielfelds
#define PIN_LED_GRUEN 9           // Launenanzeige
#define PIN_LED_GELB 8            // Launenanzeige
#define PIN_LED_ROT 7             // Launenanzeige

#define PINGEWINNTSTARTKNOPF A2   // Pin fuer den Knopf zum Starten und Beenden eines Spiels
#define PINSONGSTARTKNOPF A5      // Pin zum Starten und Beenden des Songs


SoftwareSerial vonGewinntBot(PINGEWINNTRX,PINGEWINNTTX);

Servo myservo;
int laune = 500;                  // Wert zwischen 0 und 1000
bool songSpielt = false;          // wird der Song gerade gespielt?
bool vierGewinntSpielt = false;   // läuft gerade ein Viergewinnt-Spiel?
int erwarteterEinwurf;            // speichert den Tipp, bzw. den Einwurf des Arduinos, bis dieser ausgefuehrt wird

void setup() {

  myservo.attach(PIN_SERVO);
  pinMode(PIN_LED_ROT,OUTPUT);
  pinMode(PIN_LED_GELB,OUTPUT);
  pinMode(PIN_LED_GRUEN,OUTPUT);
  
  pinMode(PINGEWINNTSTART,OUTPUT);
  pinMode(PINSONGSTART,OUTPUT);
  
  pinMode(PINGEWINNTSTARTKNOPF,INPUT);
  pinMode(PINSONGSTARTKNOPF,INPUT);

  Serial.begin(9600);
  vonGewinntBot.begin(9600);

  digitalWrite(PINSONGSTART, songSpielt);

  randomSeed(analogRead(A6));
  Serial.println("BrainBot gestartet!");
}

void loop() {


  if(laune == 0){                               // Laune zu schlecht -> "abschalten"
    gewinntBeenden();
    arduinoBeenden();
      
  }

  setzeLEDFarben();                             // Laune ueber LEDs anzeigen
  
  // Ueberpruefe Knoepfe
  if(digitalRead(PINSONGSTARTKNOPF)){           // Detektion eines Knopfdrucks vom Song-Knopf
    songSpielt = !songSpielt;
    digitalWrite(PINSONGSTART, songSpielt);
    while(digitalRead(PINSONGSTARTKNOPF));
    delay(50);
    Serial.print("Song spielt: ");
    Serial.println(songSpielt);
  }
  if(digitalRead(PINGEWINNTSTARTKNOPF)){        // Detektion eines Knopfdrucks vom Spiel-Knopf
    vierGewinntSpielt = !vierGewinntSpielt;
    while(digitalRead(PINGEWINNTSTARTKNOPF));
    delay(50);
    if (vierGewinntSpielt)
      gewinntStarten();
    else
      gewinntBeenden();
  }

                                                // Empfangen moeglicher Nachrichten des Viergewinnt-Bots
  if(vierGewinntSpielt){
    if(vonGewinntBot.available()>0)empfangen();
  }

  if(songSpielt && random(0,100) < 3)         // Anheben der Laune, waehrend der Song gespielt wird
    aendereLaune(1);
}

// Empfangen von Nachrichten des 4-Gewinnt-Bots
// Fuer die genaue Kodierung siehe Kommentare bei Funktion senden() in Online
void empfangen(){
  int empf[5];
  for(int i=0;i<5;i++){                                                       // Einlesen der Nachricht
    while(vonGewinntBot.available()==0);
    empf[i]=(int)vonGewinntBot.read();
  }
  if(empf[0]==0){                            // Nachricht kodiert Zug des Arduinos
    erwarteterEinwurf = empf[1];
    sprachAusgabe("Wirf meinen Stein in die " + String(empf[1]) + " ein!");
  }
  if(empf[0]==1){                            // Nachricht kodiert Zug, der fuer den Arduino ausgefuehrt wurde
    if(empf[2] == 10){                       // Guete 10 bedeutet ein gewonnenes Spiel
      sprachAusgabe("Ich habe gewonnen");
      aendereLaune(500);
      gewinntBeenden();
      return;
    }
    if(empf[1] == erwarteterEinwurf)         // Zug wurde korrekt ausgefuehrt
      sprachAusgabe("Danke.");
    else{                                    // ein anderer Zug wurde ausgefuehrt
      aendereLaune(-400);
      sprachAusgabe("Du solltest meinen Stein in die " + String(erwarteterEinwurf) + " einwerfen, nicht in die "  + String(empf[1]) + "!");
    }
  }
  if(empf[0]==2){                            // Nachricht kodiert Zug des Spielers
    if(empf[2] == 10){                       // Guete 10 bedeutet ein gewonnenes Spiel
      sprachAusgabe("Du hast gewonnen");
      aendereLaune(-300);
      gewinntBeenden();
      return;
    }
    if (erwarteterEinwurf == empf[1]){       // Tipp wurde befolgt
      aendereLaune(100);
      sprachAusgabe("Gut, dass du den Tipp befolgt hast.");
    }
    if(empf[2] <5){
      sprachAusgabe("Schlechter Zug!");
      aendereLaune(20);
    }
    else if(empf[2] <7)
      sprachAusgabe("Der Zug war okay.");
    else{
      sprachAusgabe("Guter Zug!");
      aendereLaune(-30);
    }
  }
  if(empf[0]==3){                            // Nachricht kodiert den Tipp, der dem Spieler gegeben wird
    erwarteterEinwurf = empf[1];
    if(empf[2] <5)
      sprachAusgabe("Ich würde deinen nächsten Stein in die " + String(empf[1]) + " werfen!");
  }
}

// Startet das Vier-Gewinnt-Spiel
void gewinntStarten(){
  Serial.println("Starte 4-Gewinnt!");
  leereFeld();
  digitalWrite(PINGEWINNTSTART,HIGH);         // sagt dem 4-Gewinnt-Bot, dass ein Spiel gestartet werden soll
  while(vonGewinntBot.available()==0);        // empfaengt den zufaellig gewaehlten Startspieler
  int startSpieler=(int)vonGewinntBot.read(); 
                                              // startspieler=1: Arduino fängt an
                                              // startspieler=2: Spieler fängt an
  if(startSpieler == 1)
    sprachAusgabe("Ich fange an!");
  else
    sprachAusgabe("Du fängst an!");
}

// beendet das 4-Gewinnt-Spiel
void gewinntBeenden(){
  digitalWrite(PINGEWINNTSTART,LOW);
  sprachAusgabe("Spiel beendet");
}

// hebt, bzw senkt die Laune, sodass sie zwischen 0 und 1000 bleibt
void aendereLaune(int aenderung){
  laune += aenderung;
  if (laune < 0)
    laune = 0;
  if (laune > 1000)
    laune = 1000;
}

// "beenden" des Arduinos
void arduinoBeenden(){
  sprachAusgabe("Tschüss");
  leereFeld();
  setzeLEDFarben();
  delay(2000000);
}

// Je nach Laune leuchten unterschiedliche Kombinationen, aus roter, gelber und gruener LED
void setzeLEDFarben(){
  if(laune > 800){
    digitalWrite(PIN_LED_GRUEN, HIGH);
    digitalWrite(PIN_LED_GELB, LOW);
    digitalWrite(PIN_LED_ROT, LOW);
  } else if (laune > 600){
    digitalWrite(PIN_LED_GRUEN, HIGH);
    digitalWrite(PIN_LED_GELB, HIGH);
    digitalWrite(PIN_LED_ROT, LOW);
  } else if (laune > 400){
    digitalWrite(PIN_LED_GRUEN, LOW);
    digitalWrite(PIN_LED_GELB, HIGH);
    digitalWrite(PIN_LED_ROT, LOW);
  } else if (laune > 200){
    digitalWrite(PIN_LED_GRUEN, LOW);
    digitalWrite(PIN_LED_GELB, HIGH);
    digitalWrite(PIN_LED_ROT, HIGH);
  } else {
    digitalWrite(PIN_LED_GRUEN, LOW);
    digitalWrite(PIN_LED_GELB, LOW);
    digitalWrite(PIN_LED_ROT, HIGH);
  }
}

// steuert den Servo, um das Spielfeld zu leeren
void leereFeld(){
  int pos = 0;
  
  for (pos = 100; pos <= 180; pos += 1) {
    // in steps of 1 degree
    myservo.write(pos);       
    delay(15);                    
  }
  for (int i=0;i<7;i++) {
    myservo.write(140);          
    delay(100);                
    myservo.write(180);
    delay(100);
  }
  for (pos = 180; pos >= 100; pos -= 1) { 
    myservo.write(pos);            
    delay(15);                      
  }
}

// gibt Strings ueber Lautsprecher aus
void sprachAusgabe(String string){
  // TODO
  Serial.println(string);
}

