
int piezoPin = 5;     // Pin des Piezos
int bassPin = 2;      // Pin zur Kommunikation mit der zweiten Stimme
int bassBeatPin = 3;  // Pin zur Mitteilung des Taktes an die zweite Stimme
int brainBotPin = 12; // Pin zur Kommunikation mit dem BrainBot


// Array zu Speicherung der Frequenzen der Tonleiter
float frequencies[] = {164.814,174.614,184.997,195.998,207.652,220.000,233.082,246.942,261.626,277.183,293.665,311.127,329.628,349.228,369.994,391.995,415.305,440.000,466.164,493.883,523.251,554.365,587.330,622.254,659.255,698.456,739.989,783.991,830.609,880.000,932.328,987.767,1046.50,1108.73,1174.66,1244.51,1318.51,1396.91,1479.98,1567.98,1661.22,1760.00,1864.66,1975.53};

// angelehnt an Gitarrensaiten die jeweilige Stelle im Array
float* E = &frequencies[0];
float* A = &frequencies[5];
float* D = &frequencies[10];
float* G = &frequencies[15];
float* B = &frequencies[19];
float* e = &frequencies[24];
float zero = 0;
float* n = &zero;

// Klasse fuer den Song
class Song{
  bool playing = false;           // Variable dafuer, ob der Song gerade spielt
  float noteLength;               // Laenge eines Taktes
  long beatStartTime;             // Zeit, zu der der letzte Takt gestartet wurde
  int toneSwitchSignal = LOW;     // Signal fuer die zweite Stimme, dass im Takt zwischen HIGH und LOW wechselt
  long noteStartTime;             // Zeit, zu der die letze Note gestartet wurde
  int counter;                    // zaehlt wieviele Noten schon gespielt wurden
  int pin;                        // Pinnummer des Piezos
  int len;                        // Laenge der Notenabfolge
  float** notes;                  // Array aus Noten
  float* duration;                // Array der zugehoerigen Laengen der Noten
  
  public: Song(float* notes[], float duration[], int len, float bpm, int pin){
    noteLength = 60/bpm*1000/2;
    this->pin = pin;
    this->len = len;
    this->notes = notes;
    this->duration = duration;
  }

// wird zu Beginn des Songs einmal aufgerufen
  void start(){
    playing = true;
    beatStartTime = millis();
    noteStartTime = millis();
    counter = 0;
  }

// wird zum Stoppen des Songs aufgerufen
  void stopp(){
    playing = false;
    noTone(pin);
  }

// wird im jeden loop()-Durchlauf einmal aufgerufen, ueberprueft, ob die Note oder die Taktangabe gewechselt werden muss
  void play(){
    if (millis() < noteStartTime + duration[counter] * noteLength)
      playNote();
    else{
      noteStartTime = millis();
      counter++;
      if (counter >= len){
        counter = 0;
        beatStartTime = millis();
      }
      playNote();
    }

    if (millis() - beatStartTime > 4*noteLength){
      beatStartTime = millis();
      if (toneSwitchSignal == HIGH)
        toneSwitchSignal = LOW;
      else
        toneSwitchSignal = HIGH;
      
    }
    digitalWrite(bassBeatPin,toneSwitchSignal);
    Serial.println(toneSwitchSignal);
  }

// spielt die Note *notes[counter]
  private: void playNote(){
    if(*notes[counter] == 0)
      noTone(pin);
    else
      tone(pin, *notes[counter]);
  }
  
};

// Noten und Spiellaengen fuer das Lied "Barbie Girl"
Song* barbieGirl = malloc(600);
float* no[100] =  {e+4, e+0, e+4, e+9, e+5, n,   B+7, B+4, e+2, e+7, e+4, e+2, e+0, n, e+0, B+2, e+2, B+2, n, e+2, e+0, e+4, e+2, n,     e+4, e+0, e+4, e+9, e+5, n,   e+5, B+7, B+4, e+2, e+7, e+4, e+2, e+0, n, e+0, n,  e+0, B+2, e+2, B+2, n, e+2, n,  e+2, n,  e+2, e+0, e+4, e+2, n};
float  du[100] =  {1,   1,   1,   1,   2,   2,   1,   1,   1,   1,   2,   1,   1,   2, 1,   1,   2,   1,   3, 1,   1,   2,   1,   1,     1,   1,   1,   1,   1.5, 1.5, 1,   1,   1,   1,   1,   2,   1,   1,   1, .9 , .1, 1,   1,   2,   1,   1, .9,  .1, .9,  .1, 1,   1,   2,   1,   1};


bool playing = false;

void setup() {
  Serial.begin(9600);
  pinMode(piezoPin,OUTPUT);
  pinMode(bassPin,OUTPUT);
  pinMode(bassBeatPin,OUTPUT);
  
  pinMode(brainBotPin,INPUT);
  barbieGirl = new Song(no, du, 55, 129, piezoPin);

  delay(1000);
  startSong();
}

void loop() {

  Serial.print(digitalRead(brainBotPin));
  Serial.print("   ");
  Serial.println(playing);
  
  if(digitalRead(brainBotPin) && !playing)  // Signal vom BrainBot: Spiele den Song
    startSong();
  if(!digitalRead(brainBotPin) && playing)  // Signal vom BrainBot: Stoppe den Song
    stopSong();
  
  if (playing)
    barbieGirl->play();
  
 
}

void startSong(){
    playing = true;
    barbieGirl->start();
    digitalWrite(bassPin, HIGH);    // Mitteilung an die zweite Stimme: starte den Song
  }

void stopSong(){
    playing = false;
    barbieGirl->stopp();
    digitalWrite(bassPin, LOW);     // Mitteilung an die zweite Stimme: stoppe den Song
  }


