//Importe
#include <SoftwareSerial.h>

//Definieren von statischen Variablen als Makros

//Pins der Lichtschranken
#define PINLICHTEINS A1         //Schranke 1
#define PINLICHTZWEI A2         //Schranke 2
#define PINLICHTDREI A3         //Schranke 3
#define PINLICHTVIER A4         //Schranke 4
#define PINLICHTFUENF A5        //Schranke 5
#define PINLICHTSECHS A6        //Schranke 6
#define PINLICHTSIEBEN A7       //Schranke 7

//Pins der Kommunikation
#define PINCOMMRX 6             //Receive-Pinb der Seriellen Kommunikation
#define PINCOMMTX 5             //Transmit-Pin der Seriellen Kommunikation
#define PINCOMMEMPFANGEN 3      //an=Spielstart; aus=Spielstopp

//Setting- Variablen
#define DURCHSCHNITTLICHE_HELLIGKEIT 135   //Grenzwert der Lichtwiderstände
#define DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE 5  //Hysterese-Wert


//Spielsteuerung, KI

class Game{
  public:
    Game();                       //Konstruktor
    void stopp();                 //Reset
    void einwerfen(int,int);      //wirft Stein ein
    int zugGuete(int,int);        //Ermittelt Güteklasse für Zug 0:sehr schlecht,10:sehr gut,5:Spalte voll
    int auswahl(int*,int);        //Ermittelt bei gleich guten Zügen den taktisch besten (vielleicht auch nicht 100% den besten)
    boolean voll();               //true, wenn board voll ist
    void drucken();               //gibt das Spielfeld auf der Konsole aus 
  private:
    int istViererTeil(int,int,int,int);     //interne Funktion, überprüft, ob der Vierer Teil des Spielbrettes ist
    int viererPruefen(int,int,int,int,int); //interne Funktion, überprüft, was für den Spieler bei dem Vierer möglich ist
    int untersteFreistelle(int);  //interne Funktion zum Ermitteln der untersten Freistelle der Spalte
    int board[7][6];              //Spielbrett, speichert alle Steine, 0 für kein Stein, 1 für Arduino, 2 für Spieler
};
Game::Game(){//Konstruktor, initialisiert alle Variablen
  for(int i=0;i<7;i++)for(int z=0;z<6;z++)this->board[i][z]=0;
}
void Game::stopp(){//resettet Spiel
  for(int i=0;i<7;i++)for(int z=0;z<6;z++)this->board[i][z]=0;
}

void Game::drucken(){
//  for(int i=0;i<6;i++){
//    for(int z=0;z<7;z++){
//      Serial.print(this->board[z][5-i]);
//      Serial.print(" ");
//    }
//    Serial.println();
//  }

  byte i, j;
  for (i = 0; i < 6; i++){
    Serial.print(i);
    for (j = 0; j < 7; j++){
      if (board[j][5-i] == 0)
        Serial.print("  ");
      if (board[j][5-i] == 1)
        Serial.print(" X");
      if (board[j][5-i] == 2)
        Serial.print(" O");
    }
    Serial.print("\n");
  }
  Serial.println("  1 2 3 4 5 6 7");
}

int Game::zugGuete(int col,int player){//Ermittelt Güteklasse für Zug
  //gewinnen:                     10
  //verlieren vereiteln:          9
  //Zwickmühle vereiteln:         8
  //Zwickmühle aufbauen:          7
  //auf Zwickmühle reingefallen:  6
  //nichts relevantes:            5
  //Sieg womöglich verbauen:      3
  //Zwickmühle ermöglichen:       2
  //verlieren ermöglichen:        1
  //Spalte voll:                  0
  if(this->board[col][5]!=0)return 0;
  int* viererSpieler=new int[16];//Liste der Möglichen Vierer und Dreier beider Spieler auf dem Feld in das eingeworfen wird
  int* viererGegner=new int[16];//Liste der Möglichen Vierer und Dreier beider Spieler auf dem Feld über dem in das eingeworfen wird
  int stelle=untersteFreistelle(col);
  for(int i=0;i<4;i++)viererSpieler[i]=viererPruefen(col,stelle+i,0,-1,player);
  for(int i=0;i<4;i++)viererSpieler[i+4]=viererPruefen(col+i,stelle,-1,0,player);
  for(int i=0;i<4;i++)viererSpieler[i+8]=viererPruefen(col+i,stelle-i,-1,1,player);
  for(int i=0;i<4;i++)viererSpieler[i+12]=viererPruefen(col+i,stelle+i,-1,-1,player);
  for(int i=0;i<4;i++)viererGegner[i]=viererPruefen(col,stelle+1+i,0,-1,3-player);
  for(int i=0;i<4;i++)viererGegner[i+4]=viererPruefen(col+i,stelle+1,-1,0,3-player);
  for(int i=0;i<4;i++)viererGegner[i+8]=viererPruefen(col+i,stelle+1-i,-1,1,3-player);
  for(int i=0;i<4;i++)viererGegner[i+12]=viererPruefen(col+i,stelle+1+i,-1,-1,3-player);
  for(int i=0;i<16;i++)if(viererSpieler[i]==10)return 10;//Spieler gewinnt durch Einwurf
  int numGegnerVierAlt=0;
  int numSpielerDreiAlt=0;
  int numGegnerDreiAlt=0;
  int numSpielerVierNeu=0;
  int numGegnerVierNeu=0;
  int numSpielerDreiNeu=0;
  int numGegnerDreiNeu=0;
  for(int i=0;i<16;i++){
    if(viererSpieler[i]==9)numGegnerVierAlt++;
    if(viererSpieler[i]==8)numSpielerDreiAlt++;
    if(viererSpieler[i]==7)numGegnerDreiAlt++;
    if(viererGegner[i]==10)numGegnerVierNeu++;
    if(viererGegner[i]==9)numSpielerVierNeu++;
    if(viererGegner[i]==8)numGegnerDreiNeu++;
    if(viererGegner[i]==7)numSpielerDreiNeu++;
  }
  delete[](viererSpieler);//Alle Informationen mit Relevanz extrahiert
  delete[](viererGegner); //---------||-------------------------------
  viererSpieler=nullptr;
  viererGegner=nullptr;
  if(numGegnerVierNeu>0&&numGegnerVierAlt>0)return 6;//auf Zwickmühle reingefallen
  if(numGegnerVierNeu>0)return 1;//verlieren ermöglichen
  if(numGegnerVierAlt>0)return 9;//verlieren vereiteln
  if(numSpielerDreiAlt>1)return 7;//Zwickmühle aufbauen
  if(numSpielerDreiAlt>0&&numSpielerVierNeu>0)return 7;//Zwickmühle aufbauen
  if(numGegnerDreiAlt>1)return 8;//Zwickmühle vereiteln
  if(numGegnerDreiNeu>1)return 2;//Zwickmühle ermöglichen
  if(numSpielerVierNeu>0)return 3;//Sieg womöglich verbauen
  return 5;//nichts relevantes
}
int Game::viererPruefen(int w,int h,int dirw,int dirh,int player){
  if(!((w<7&&w>=0&&h<6&&h>=0)&&((w+3*dirw)<7&&(w+3*dirw)>=0&&(h+3*dirh)<6&&(h+3*dirh)>=0)))return 11;//Vierer liegt nicht komplett im Feld, Irrelevant
  int anzahlSpieler=0;
  int anzahlGegner=0;
  for(int i=0;i<4;i++){
    int x=this->board[w+i*dirw][h+i*dirh];
    if(x==player)anzahlSpieler++;
    else if(x==0);
    else anzahlGegner++;
  }
  if(anzahlSpieler+anzahlGegner>3)return 11; //FEHLER, sollte nicht passieren
  if(anzahlSpieler==3)return 10;//Spieler kann hier 4er bauen
  if(anzahlGegner==3)return 9;//Gegner kann hier 4er bauen
  if(anzahlSpieler==2&&anzahlGegner==0)return 8;//Spieler kann hier 3er bauen
  if(anzahlGegner==2&&anzahlSpieler==0)return 7;//Gegner kann hier 3er bauen
  return 6;//Gemischte Steine oder nur 1 Stein im Feld
}
void Game::einwerfen(int col,int player){//wirft Stein für Spieler player in der Spalte col ein, col darf nicht voll sein!
  this->board[col][untersteFreistelle(col)]=player;
}
int Game::untersteFreistelle(int col){//ermittelt untesten Freiplatz für Stein, col darf nicht voll sein!
  for(int i=5;i>0;i--){
    if(this->board[col][i-1]!=0)return i;
  }
  return 0;
}
int Game::istViererTeil(int w,int h,int dirw,int dirh){
  if(!((w<7&&w>=0&&h<6&&h>=0)&&((w+3*dirw)<7&&(w+3*dirw)>=0&&(h+3*dirh)<6&&(h+3*dirh)>=0)))return 0;
  return 1;
}
int Game::auswahl(int* cols,int l){
  int* newcols=new int[l];
  int* anzahlVierer=new int[l];
  int z=0;
  for(int i=0;i<7;i++){
    if(cols[i]!=11){
      newcols[z]=cols[i];
      z++;
    }
  }
  for(int i=0;i<l;i++){
    int stelle=untersteFreistelle(newcols[i]);
    for(int z=0;z<4;z++)anzahlVierer[i]+=istViererTeil(newcols[i],stelle+z,0,-1);
    for(int z=0;z<4;z++)anzahlVierer[i]+=istViererTeil(newcols[i]+z,stelle,-1,0);
    for(int z=0;z<4;z++)anzahlVierer[i]+=istViererTeil(newcols[i]+z,stelle-z,-1,1);
    for(int z=0;z<4;z++)anzahlVierer[i]+=istViererTeil(newcols[i]+z,stelle+z,-1,-1);
  }
  int maximum=0;
  for(int i=0;i<l;i++)if(anzahlVierer[i]>maximum)maximum=anzahlVierer[i];
  for(int i=0;i<l;i++)if(maximum-anzahlVierer[i]>7-(l/2))anzahlVierer[i]=0;
  while(true){
    int a=random(0,l);
    if(anzahlVierer[a]!=0)return newcols[a];
  }
}
boolean Game::voll(){
  return ((this->board[0][5]!=0)&&(this->board[1][5]!=0)&&(this->board[2][5]!=0)&&(this->board[3][5]!=0)&&(this->board[4][5]!=0)&&(this->board[5][5]!=0)&&(this->board[6][5]!=0));
}

//Arduino

bool playing; //true, wenn Spiel aktiv
int player;  //1, wenn Arduino spielt, 2 für Spieler
Game* spiel;  //Variable für Spiel
SoftwareSerial anBrainBot(PINCOMMRX,PINCOMMTX);

void setup() {
  Serial.begin(9600);
  randomSeed(analogRead(9));
  pinMode(PINLICHTEINS,INPUT);
  pinMode(PINLICHTZWEI,INPUT);
  pinMode(PINLICHTDREI,INPUT);
  pinMode(PINLICHTVIER,INPUT);
  pinMode(PINLICHTFUENF,INPUT);
  pinMode(PINLICHTSECHS,INPUT);
  pinMode(PINLICHTSIEBEN,INPUT);
  pinMode(PINCOMMEMPFANGEN,INPUT);
  anBrainBot.begin(9600);
  playing=false;
  player=0;
  spiel=new Game();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void loop() {
  if(digitalRead(PINCOMMEMPFANGEN)==HIGH && !playing)startGame();
  if (digitalRead(PINCOMMEMPFANGEN)==LOW && playing)stopGame();
  if(player==2&&playing){//Status: Arduino wartet auf Einwurf eines Steines, Spiel läuft
    Serial.println("Warte auf Einwurf fuer Spieler");
    spiel->drucken();
    int eingeworfen = warteAufEinwurf();
    if(eingeworfen == 7)
      return;
    einwurfGegner(eingeworfen);
  }
  else if(playing){//Status: Arduino ist dran, Spiel läuft
    int* arr=new int[7];
    for(int i=0;i<7;i++){
      arr[i]=spiel->zugGuete(i,1);
    }
    int slot;
    int best=0;
    int zahl=0;
    for(int i=0;i<7;i++)if(arr[i]>best)best=arr[i];
    for(int i=0;i<7;i++){
      if(arr[i]!=best)arr[i]=11;
      else zahl++;
    }
    for(int i=0;i<7;i++)if(arr[i]!=11)arr[i]=i;
    slot=spiel->auswahl(arr,zahl);
    delete[](arr);
    senden(0,slot,best,zahl,0);
    Serial.print("Warte auf Einwurf von Arduino in die ");
    Serial.println(slot+1);
    spiel->drucken();
    int eingeworfen = warteAufEinwurf();
    if(eingeworfen == 7)
      return;
    senden(1,eingeworfen,spiel->zugGuete(eingeworfen,1),0,unentschieden());
    spiel->einwerfen(eingeworfen,1);
    player=2;
    Serial.print("Der Zug hatte die Guete: ");
    Serial.println(best);
    if(best==10||spiel->voll()){
      stopGame();
    }
    else tipp();
  }
}


int warteAufEinwurf(){        // wartet auf den Einwurf eines Spielsteins und gibt die entsprechende Spalte zurueck
    int eingeworfen=7;
    int verzoegerung = 1000;
    while(eingeworfen==7){
      if(digitalRead(PINCOMMEMPFANGEN)==LOW){
        stopGame();
        break;
      }
      if(analogRead(PINLICHTEINS)<DURCHSCHNITTLICHE_HELLIGKEIT){
        eingeworfen=0;
        delay(verzoegerung);
        while(analogRead(PINLICHTEINS)<DURCHSCHNITTLICHE_HELLIGKEIT + DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE);
      }
      if(analogRead(PINLICHTZWEI)<DURCHSCHNITTLICHE_HELLIGKEIT){
        eingeworfen=1;
        delay(verzoegerung);
        while(analogRead(PINLICHTZWEI)<DURCHSCHNITTLICHE_HELLIGKEIT + DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE);
      }
      if(analogRead(PINLICHTDREI)<DURCHSCHNITTLICHE_HELLIGKEIT){
        eingeworfen=2;
        delay(verzoegerung);
        while(analogRead(PINLICHTDREI)<DURCHSCHNITTLICHE_HELLIGKEIT + DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE);
      }
      if(analogRead(PINLICHTVIER)<DURCHSCHNITTLICHE_HELLIGKEIT){
        eingeworfen=3;
        delay(verzoegerung);
        while(analogRead(PINLICHTVIER)<DURCHSCHNITTLICHE_HELLIGKEIT + DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE);
      }
      if(analogRead(PINLICHTFUENF)<DURCHSCHNITTLICHE_HELLIGKEIT){
        eingeworfen=4;
        delay(verzoegerung);
        while(analogRead(PINLICHTFUENF)<DURCHSCHNITTLICHE_HELLIGKEIT + DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE);
      }
      if(analogRead(PINLICHTSECHS)<DURCHSCHNITTLICHE_HELLIGKEIT){
        eingeworfen=5;
        delay(verzoegerung);
        while(analogRead(PINLICHTSECHS)<DURCHSCHNITTLICHE_HELLIGKEIT + DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE);
      }
      if(analogRead(PINLICHTSIEBEN)<DURCHSCHNITTLICHE_HELLIGKEIT){
        eingeworfen=6;
        delay(verzoegerung);
        Serial.println("TEST");
        while(analogRead(PINLICHTSIEBEN)<DURCHSCHNITTLICHE_HELLIGKEIT + DURCHSCHNITTLICHE_HELLIGKEIT_HYSTERESE) Serial.println(analogRead(PINLICHTSIEBEN));
      }
    }
    if (eingeworfen != 8){
      Serial.print("Einwurf in der ");
      Serial.println(eingeworfen+1);
    }
    return eingeworfen;
}

void startGame(){//Startet das Spiel mit zufälligem Startspieler
  Serial.println("Spiel gestartet");
  if(!playing){
    playing=true;
    player=random(1,3);//Zufallszahl zwischen 1 und 2 für beginnenden Spieler
    if(player==1)anBrainBot.write((char)1);
    else {
      anBrainBot.write((char)2);
      tipp();
    }
  }
}

void stopGame(){//Stoppt das Spiel
  Serial.println("Spiel beendet");
  if(playing){
    playing=false;
    spiel->stopp();
  }
}

void tipp(){
  int arr[7];
  for(int i=0;i<7;i++){
    arr[i]=spiel->zugGuete(i,2);
  }
  int slot=-1;
  int worst=12;
  for(int i=0;i<7;i++)if(arr[i]<worst&&arr[i]>0)worst=arr[i];
  while(slot==-1){
    int i=random(0,7);
    if(arr[i]==worst)slot=i;
  }
  senden(3,slot,0,0,0);
  Serial.print("Tipp :");
  Serial.println(slot+1);
}

void einwurfGegner(int col){
  int guete=spiel->zugGuete(col,2);
  spiel->einwerfen(col,2);
  player=1;
  senden(2,col,guete,0,unentschieden());
  if(guete==10||spiel->voll()){
    stopGame();
  }
  delay(100);
}

void senden(int typ,int zug,int guete,int anzahl,int unentschieden){//typ: 0-3 steht für Typunterschied der Sendung; zug: 0-6 Spalte; guete: 1-10 Zuggüte; anzahl:1-7 Anzahl der Spalten mit guete; unentschieden: 0-1 1=unentschieden
                                      //Typ 0: Arduinos Zug
                                      //Typ 1: Einwurf des Spielers für den Arduino
                                      //Typ 2: Bewertung des Spielerzugs
                                      //Typ 3: Tipp für Spieler
  if(typ<0||typ>3||zug<0||zug>6||guete<0||guete>10||unentschieden<0||unentschieden>1)return;
  anBrainBot.write((char)typ);
  anBrainBot.write((char)zug+1);
  anBrainBot.write((char)guete);
  anBrainBot.write((char)anzahl);
  anBrainBot.write((char)unentschieden);
}

int unentschieden(){
  if(spiel->voll())return 1;
  else return 0;
}
